'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.mapForReact = mapForReact;
exports.mapToRelative = mapToRelative;
exports.mapModule = mapModule;
var path = require('path');

var fs = require('fs');

var dirContentsMap = {};
var regExpMap = {
  react: {
    regexp: new RegExp('auto:', 'g'),
    replace: './'
  },
  npm: {
    regexp: new RegExp('npm:', 'g'),
    replace: ''
  }
};

var lastIn = void 0;
var lastOut = void 0;

function getStateOptions(s) {
  if (s.opts) {
    if (Array.isArray(s.opts)) {
      return { map: s.opts };
    } else if (!s.opts.map && s.opts.expose) {
      return { map: [s.opts] };
    }

    return s.opts;
  }

  return {};
}

function getRootPath(o) {
  // throw `IGNORE_ABSOLUTE: ${process.env.IGNORE_ABSOLUTE} | root: ${o.root} | ${JSON.stringify(o, null, 2)} | abs: ${path.isAbsolute(o.root || '')}`;
  var root = o.root === './' ? path.resolve(__dirname, '..') : o.root;
  return !process.env.IGNORE_ABSOLUTE && root ? root : '';
}

function reactOsFileInfer(s) {
  return s && s.react;
}

function escapeRegExp(str) {
  return str.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, '\\$&');
}

function createFilesMap(options) {
  var keys = [];
  var contents = {};
  var map = options.map || [];

  map.forEach(function (m) {
    var expose = m.expose || undefined;
    var src = m.src || undefined;

    if (expose) {
      if (typeof src === 'string') {
        // throw `root ${rootPath} | src: ${src}`;
        // (^config[\/]+|[\/]+config$|[\/]+config[\/]+)

        var escapedExpose = escapeRegExp(expose);

        keys.push(expose);
        contents[expose] = {
          expose: expose,
          regexp: new RegExp('(^' + escapedExpose + '$)|(^' + escapedExpose + '[/]+)|([/]+' + escapedExpose + '[/]+)|([/]+' + escapedExpose + ')'),
          value: src
        };
      }
    }

    return null;
  });

  return { keys: keys, contents: contents };
}

function resolve(filename) {
  if (path.isAbsolute(filename)) return filename;
  return path.resolve(process.cwd(), filename);
}

function toPosixPath(modulePath) {
  return modulePath.replace(/\\/g, '/');
}

function toReactPath(base, fileList, o) {
  // TODO: Check if the current file is a directory, if it is check the logic on index.js
  var os = (o || process.env.TARGET_PLATFORM || '').toLowerCase();

  // throw new Error(`${os} | ${base} | ${fileList}`);
  // Check for mobile substitutions first
  if (['mobile', 'ios', 'android', 'windows'].indexOf(os) > -1) {
    // Check for a mobile file first
    if (fileList.indexOf(base + '.mobile.js') > -1) {
      return base + '.mobile';
    }

    // Check for OS specific file
    if (fileList.indexOf(base + '.' + os + '.js') > -1) {
      return base + '.' + os;
    }

    // Fallback on normal file (web/desktop) for passthrow files
    return toReactPath(base, fileList, 'desktop');
  } else if (os === 'desktop') {
    // Check for desktop only files
    if (fileList.indexOf(base + '.desktop.js') > -1) {
      return base + '.desktop';
    }

    // Fallback on web
    return toReactPath(base, fileList, 'web');
  } else if (os === 'web') {
    if (fileList.indexOf(base + '.web.js') > -1) {
      return base + '.web';
    }
  }

  // Suffixless files
  if (fileList.indexOf(base + '.js') > -1) {
    return '' + base;
  }

  return base;
}

function mapForReact(moduleMapped) {
  // moduleMapped =
  var base = path.dirname(moduleMapped);

  // Index files in destination directory once
  if (!dirContentsMap[base]) {
    dirContentsMap[base] = fs.readdirSync(base).map(function (v) {
      return v.toLowerCase();
    });
  }

  // Fix mapped module path
  var newFile = toReactPath(path.basename(moduleMapped), dirContentsMap[base]);

  return base + '/' + newFile;
}

function determineContext(fileName, options) {
  var fileData = {
    file: '',
    react: false,
    npm: false
  };

  var arg = fileName;

  Object.keys(regExpMap).forEach(function (k) {
    if (arg.search(regExpMap[k].regexp) > -1) {
      arg = arg.replace(regExpMap[k].regexp, regExpMap[k].replace);

      if (arg === 'unknown') {
        arg = '';
      } else {
        fileData[k] = true;
      }
    }
  });

  if (arg && arg !== 'unknown') {
    fileData.file = arg;
  }

  fileData.react = fileData.react && options.react;
  return fileData;
}

// Rewrite module path relativelly
function mapToRelative(moduleFile, constantModulePart, options) {
  var from = resolve(path.dirname(constantModulePart));
  var to = resolve(path.normalize(moduleFile));
  var moduleMapped = toPosixPath(path.relative(from, to));

  // throw JSON.stringify({
  //   from,
  //   to,
  //   moduleMapped,
  //   moduleFile,
  //   constantModulePart,
  //   options
  // }, null, 2);

  // Support React-Native specific require rewrites
  if (reactOsFileInfer(options)) {
    return mapForReact(moduleMapped);
  }

  return moduleMapped;
}

// Adapts a module path to the context of caller
function adaptModulePath(modulePath, state) {
  // const fileName = state.file.opts.filename;
  var options = getStateOptions(state);
  var filesMap = createFilesMap(options);
  var rootPath = getRootPath(options);

  var module = determineContext(modulePath, options);

  // Do not generate infinite cyrcular references on empty nodes
  if (!module.file) {
    return null;
  }

  // Safeguard against circular calls
  if (lastIn === lastOut) {
    return null;
  }

  // Remove relative path prefix before replace
  var absoluteModule = path.isAbsolute(module.file);

  // Try to replace aliased module
  var found = false;
  var constantModulePart = void 0;
  filesMap.keys.filter(function (k) {
    var d = filesMap.contents[k];
    var idx = module.file.search(d.regexp);

    if (!found && idx > -1) {
      // throw `Found ${d} in ${module.file}`
      // constantModulePart = module.file.slice(0, idx) || './';
      constantModulePart = './';

      if (module.file[idx] === '/') {
        idx += 1;
      }

      var value = d.value[0] === '.' && idx > 0 ? d.value.slice(2) : d.value;
      // const value = d.value;

      // Replace the alias with it's path and continue
      module.file = '' + (module.file.slice(0, idx) || '') + value + (module.file.slice(idx + d.expose.length) || '');
      found = true;

      // Revaluate npm and react flags based on the new mapping
      module = determineContext(module.file, options);

      return true;
    }

    return false;
  });

  // Leave NPM modules as resolved, do not remap and ignore wrongfully formatted strings of form require('npm:')
  if (module.npm) {
    return module.file || null;
  }

  // Do not touch direct requires to npm modules (non annotated)
  if (module.file.indexOf('./') !== 0 && module.file.indexOf('/') !== 0) {
    return null;
  }

  // Check if any substitution took place
  if (found) {
    // Module alias substituted
    if (!path.isAbsolute(module.file)) {
      if (!absoluteModule && module.file[0] !== '.') {
        // Add the relative notation back
        module.file = './' + module.file;
      }
    }
  }

  // Do not break node_modules required by name
  if (!found && module.file[0] !== '.') {
    if (reactOsFileInfer(options)) {
      var aux2 = mapForReact(module.file);

      if (aux2 !== module.file) {
        return aux2;
      }
    }

    return null;
  }

  // Enforce absolute paths on absolute mode
  if (rootPath) {
    if (!path.isAbsolute(module.file)) {
      module.file = path.join(rootPath, module.file);

      if (reactOsFileInfer(options)) {
        return mapForReact(module.file);
      }

      return module.file;
    }

    // After the node is replaced the visitor will be called again.
    // Without this condition these functions will generate a circular loop.
    return null;
  }

  // throw `Gets here: ${JSON.stringify(module.file)}`;

  // Do not bother with relative paths that are not aliased
  if (!found) {
    return null;
  }

  var moduleMapped = mapToRelative(module.file, constantModulePart, options);
  if (moduleMapped.indexOf('./') !== 0) {
    moduleMapped = './' + moduleMapped;
  }

  // throw JSON.stringify({
  //   moduleMapped, modulePath, module, constantModulePart
  // }, null, 2);
  return moduleMapped !== modulePath ? moduleMapped : null;
}

// Safeguard passthrow function remembers last input and generated output
function mapModule(modulePath, state) {
  lastIn = '' + modulePath + JSON.stringify(state) + process.env.IGNORE_ABSOLUTE;
  var out = adaptModulePath(modulePath, state);
  lastOut = '' + out + JSON.stringify(state) + process.env.IGNORE_ABSOLUTE;

  // if (out === './test/webpack.dev.config/.config') {
  //   throw `called again`
  // }

  return out;
}

exports.default = function (_ref) {
  var t = _ref.types;

  function transformRequireCall(nodePath, state) {
    if (!t.isIdentifier(nodePath.node.callee, { name: 'require' }) && !(t.isMemberExpression(nodePath.node.callee) && t.isIdentifier(nodePath.node.callee.object, { name: 'require' }))) {
      return;
    }

    var moduleArg = nodePath.node.arguments[0];
    if (moduleArg && moduleArg.type === 'StringLiteral') {
      var modulePath = mapModule(moduleArg.value, state);

      if (modulePath) {
        nodePath.replaceWith(t.callExpression(nodePath.node.callee, [t.stringLiteral(modulePath)]));
      }
    }
  }

  function transformImportCall(nodePath, state) {
    var moduleArg = nodePath.node.source;
    if (moduleArg && moduleArg.type === 'StringLiteral') {
      var modulePath = mapModule(moduleArg.value, state);

      if (modulePath) {
        nodePath.replaceWith(t.importDeclaration(nodePath.node.specifiers, t.stringLiteral(modulePath)));
      }
    }
  }

  return {
    visitor: {
      CallExpression: {
        exit: function exit(nodePath, state) {
          return transformRequireCall(nodePath, state);
        }
      },
      ImportDeclaration: {
        exit: function exit(nodePath, state) {
          return transformImportCall(nodePath, state);
        }
      }
    }
  };
};